/*
 * Written by Rymma from the Phoenix world on April 7th, 2009.
 * Proud Bastoker and member of the BlueIron linkshell.
 * 
 * MMMRunePlacer is the main UI of this application. It also creates the available shapes and the mazes, and controls everything.
 * Some of the UI event is delegated to Maze.
 * Shape is an Icon.
 * Maze is a JPanel (for mouse events).
 * 
 * A solving thread runs and iterates the algorithm step by step and updates the screen & the loading bar.
 */

package rymma.mmm;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class MMMRunePlacer extends JPanel implements ChangeListener, Runnable, ActionListener {

	// Main (Windowed)
	public static void main(String[] args) {
		JFrame window = new JFrame();
		window.setTitle("Rymma's Moblin Maze Mongers Rune Placer");
		window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		window.setContentPane(new MMMRunePlacer());
		window.setBounds(100, 100, 1000, 800);
		window.setVisible(true);
	}

	// MMMRunePlacer
	public Map<Shape, JSpinner> shapeSpinners;

	public List<Shape> shapes;

	public Maze maze;

	public static final int STATE_SOLVED = 0;

	public static final int STATE_END = 1;

	public static final int STATE_BUST = 2;

	public static final int STATE_RUNNING = 3;

	public int state;

	public static final String[] stateString = { "Solved ^_^", "Reached end, sorry -_-", "You have more blocks than free maze cell -_-", "Thinking ^_^" };

	public JLabel status;

	public JButton resume;

	public JTextField refreshInterval;

	public MMMRunePlacer() {
		super(new BorderLayout());

		// define types of rune shape
		Shape shape_1x2 = new Shape(new boolean[][] { { true, true } }, 2, "1x2");
		Shape shape_1x3 = new Shape(new boolean[][] { { true, true, true } }, 2, "1x3");
		Shape shape_1x4 = new Shape(new boolean[][] { { true, true, true, true } }, 2, "1x4");
		Shape shape_2x2 = new Shape(new boolean[][] { { true, true }, { true, true } }, 1, "2x2");
		Shape shape_S = new Shape(new boolean[][] { { false, true, true }, { true, true, false } }, 2, "S");
		Shape shape_Z = new Shape(new boolean[][] { { true, true, false }, { false, true, true } }, 2, "Z");
		Shape shape_cross = new Shape(new boolean[][] { { false, true, false }, { true, true, true }, { false, true, false } }, 1, "Cross");
		Shape shape_corner = new Shape(new boolean[][] { { true, true }, { true, false } }, 4, "Corner");
		Shape shape_T = new Shape(new boolean[][] { { false, true, false }, { true, true, true } }, 4, "T");
		Shape shape_L = new Shape(new boolean[][] { { true, true }, { true, false }, { true, false } }, 4, "L");
		Shape shape_J = new Shape(new boolean[][] { { true, false }, { true, false }, { true, true } }, 4, "J");

		// register each shape in a list
		// I'm implicitly adding the shapes in decresing order of their block size, so that search is speed up
		this.shapes = new LinkedList<Shape>();
		shapes.add(shape_1x4);
		shapes.add(shape_2x2);
		shapes.add(shape_S);
		shapes.add(shape_Z);
		shapes.add(shape_T);
		shapes.add(shape_L);
		shapes.add(shape_J);
		shapes.add(shape_cross);
		shapes.add(shape_corner);
		shapes.add(shape_1x3);
		shapes.add(shape_1x2);

		// create rune shape list with counters
		shapeSpinners = new HashMap<Shape, JSpinner>();
		JPanel counters = new JPanel(new GridLayout(shapes.size(), 2));
		for (Shape shape : shapes) {
			counters.add(new JLabel(shape));
			JSpinner spinner = new JSpinner(new SpinnerNumberModel(0, 0, 10, 1));
			spinner.addChangeListener(this);
			shapeSpinners.put(shape, spinner);
			counters.add(spinner);
		}
		this.add(new JScrollPane(counters), BorderLayout.EAST);

		// controls
		resume = new JButton("Find another solution");
		refreshInterval = new JTextField("0", 4);
		JLabel refreshIntervalText = new JLabel("Refresh rate");
		JPanel controls = new JPanel();
		controls.add(resume);
		//controls.add(refreshIntervalText);
		//controls.add(refreshInterval);
		JPanel mazeAndControls = new JPanel(new BorderLayout());

		resume.addActionListener(this);

		// create maze
		this.maze = new Maze(this);
		mazeAndControls.add(new JScrollPane(controls), BorderLayout.SOUTH);
		mazeAndControls.add(new JScrollPane(maze), BorderLayout.CENTER);
		this.add(mazeAndControls, BorderLayout.CENTER);

		// status bar
		this.status = new JLabel("");
		this.add(status, BorderLayout.SOUTH);
		setState(STATE_SOLVED);

		// top section with text
		JPanel top = new JPanel(new GridLayout(4, 1));
		top.add(new JLabel("Written by Rymma from the Phoenix world on April 7th, 2009."));
		top.add(new JLabel("Proud Bastoker and member of the BlueIron linkshell."));
		top.add(new JLabel("Click on grid to add blocking cells. Add rune shapes to the right and look at bottom status bar for result."));
		top.add(new JLabel("Version 1.0"));
		this.add(top, BorderLayout.NORTH);

		// run
		Thread t = new Thread(this);
		t.start();
	}

	public void setState(int state) {
		synchronized (this) {
			this.state = state;
			this.status.setText(stateString[state]);
			resume.setEnabled(state == STATE_SOLVED);
		}
	}

	// ChangeListener
	public void stateChanged(ChangeEvent e) {
		// get value of each spinner, clear the current shapes, and add all the shapes again
		// Note: I'm implicitly adding the shapes in decresing order of their block size, so that search is speed up (see constructor)
		synchronized (this) {
			setState(STATE_SOLVED);
			maze.removeAllShapes();
			for (Shape shape : shapes) {
				JSpinner spinner = shapeSpinners.get(shape);
				int n = ((Integer) spinner.getModel().getValue()).intValue();
				for (int i = 0; i < n; i++) {
					maze.addShape(shape);
				}
			}
			if (maze.isBustingPieceCount()) {
				setState(STATE_BUST);
			} else {
				setState(STATE_RUNNING);
			}
		}
	}

	// ActionListener
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == resume) {
			setState(STATE_RUNNING);
		}
	}

	// Runnable
	public void run() {
		try {
			long snap = System.currentTimeMillis();
			int interval = 0;
			while (true) {
				Thread.yield();
				if (state == STATE_RUNNING) {
					synchronized (this) {
						maze.step();
						/*
						 try {
						 interval = Integer.parseInt(refreshInterval.getText());
						 } catch (Exception e) {
						 interval = 0;
						 }
						 */
						if (maze.isSolution()) {
							setState(STATE_SOLVED);
						} else if (maze.isLastStep()) {
							setState(STATE_END);
						}
						long now = System.currentTimeMillis();
						if (now > snap + interval || state != STATE_RUNNING) {
							maze.refresh();
							snap = now;
						}
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
