/*
 * Written by Rymma from the Phoenix world on April 7th, 2009.
 * Proud Bastoker and member of the BlueIron linkshell.
 * 
 * Maze describes a maze with pits.
 * 
 * It's also a JPanel to support mouse events.
 * 
 * TODO support elemental
 */

package rymma.mmm;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JPanel;
import javax.swing.event.ChangeListener;

public class Maze extends JPanel implements MouseListener {

	// Maze
	public int width;

	public int height;

	public int grid[][];

	public Color colors[];

	// the following are computation states
	public int nShape;

	public Shape shapes[];

	public int positionsR[];

	public int positionsC[];

	public int rotations[];

	public boolean placed[];

	public ChangeListener listener;

	public Maze(ChangeListener listener) {
		this.listener = listener;
		this.width = 5;
		this.height = 5;
		this.grid = new int[height][width];
		this.colors = new Color[width * height + 2];
		colors[0] = Color.WHITE; // empty
		colors[1] = Color.BLACK; // pits
		colors[2] = Color.BLUE;
		colors[3] = Color.RED;
		colors[4] = Color.GREEN;
		colors[5] = Color.CYAN;
		colors[6] = Color.MAGENTA;
		colors[7] = Color.ORANGE;
		colors[8] = Color.PINK;
		colors[9] = Color.YELLOW;
		colors[10] = Color.GRAY;
		colors[11] = Color.LIGHT_GRAY;
		for (int i = 12; i < colors.length; i++) {
			colors[i] = new Color((int) (Math.random() * 0x00FFFFFF));
		}

		this.setPreferredSize(new Dimension(width * blockSize + 1, height * blockSize + 1)); // +1 for borders
		this.addMouseListener(this);

		// init states
		this.nShape = 0;
		int capacity = width * height;
		this.shapes = new Shape[capacity];
		this.positionsR = new int[capacity];
		this.positionsC = new int[capacity];
		this.rotations = new int[capacity];
		this.placed = new boolean[capacity];
	}

	// Methods
	public synchronized void togglePit(int r, int c) {
		reset();
		if (grid[r][c] != 1) grid[r][c] = 1;
		else grid[r][c] = 0;
		refresh();
		listener.stateChanged(null);
	}

	public synchronized void addShape(Shape s) {
		if (nShape == shapes.length) return; // cannot add more shapes (does not corrupt anything since we can only removeAll and addOne)
		shapes[nShape] = s;
		nShape++;
		reset(); // to reset computing states
	}

	public synchronized void removeAllShapes() {
		nShape = 0;
		reset(); // to clear board
	}

	public synchronized void reset() {
		// reset computing states
		for (int i = 0; i < nShape; i++) {
			// mask is square, even if shape isn't, so we have to start in the negative
			positionsR[i] = -shapes[i].dimension + 1;
			positionsC[i] = -shapes[i].dimension + 1;
			rotations[i] = -1; // so that first step goes to the next rotation, which is 0
			placed[i] = false;
		}
		// clear board of pieces
		for (int r = 0; r < height; r++) {
			for (int c = 0; c < width; c++) {
				if (grid[r][c] > 1) grid[r][c] = 0;
			}
		}
	}

	public synchronized void refresh() {
		// request a repaint of this panel
		this.repaint();
	}

	public synchronized boolean isSolution() {
		boolean solution = true;
		for (int i = 0; i < nShape && solution; i++) {
			solution &= placed[i];
		}
		return solution;
	}

	public synchronized boolean isLastStep() {
		return positionsR[0] == height;
	}

	// compute a single step in the algolrithm
	public synchronized void step() {
		if (nShape == 0) return;

		// goto first unplaced rune or last rune if none
		int rune = -1;
		for (int i = 0; i < nShape && rune == -1; i++) {
			if (!placed[i]) rune = i;
		}
		if (rune == -1) rune = nShape - 1;
		int id = rune + 2; // 0 & 1 are reserved id for empty & pits

		// unset rune
		unsetRune(id, rune);

		// try next rotation
		rotations[rune]++;
		if (rotations[rune] < shapes[rune].nMask) {
			int r = positionsR[rune];
			int c = positionsC[rune];
			boolean fits = testRune(r, c, shapes[rune], rotations[rune]);
			if (fits) setRune(id, r, c, shapes[rune], rotations[rune], rune);
		} else {
			// if no other rotation exists, reset rotation state & try next position
			rotations[rune] = 0;
			positionsC[rune]++;
			if (positionsC[rune] >= width) {
				positionsC[rune] = -shapes[rune].dimension + 1;
				positionsR[rune]++;
			}
			if (positionsR[rune] < height) {
				int r = positionsR[rune];
				int c = positionsC[rune];
				boolean fits = testRune(r, c, shapes[rune], rotations[rune]);
				if (fits) setRune(id, r, c, shapes[rune], rotations[rune], rune);
			} else {
				// if no other position exists, reset position & unset previous rune
				if (rune > 0) {
					positionsR[rune] = -shapes[rune].dimension + 1;
					positionsC[rune] = -shapes[rune].dimension + 1;
					unsetRune(id - 1, rune - 1);
				}
			}
		}
	}

	public synchronized boolean testRune(int ir, int ic, Shape s, int mask) {
		// test if mask fits position
		for (int r = 0; r < s.dimension; r++) {
			for (int c = 0; c < s.dimension; c++) {
				if (s.masks[mask][r][c]) {
					if (ir + r < 0 || ic + c < 0 || ir + r >= height || ic + c >= width || grid[ir + r][ic + c] != 0) return false;
				}
			}
		}
		return true;
	}

	public synchronized void setRune(int id, int ir, int ic, Shape s, int mask, int runeIndex) {
		// sets without checks (caller is responsible to calling testRune before)
		for (int r = 0; r < s.dimension; r++) {
			for (int c = 0; c < s.dimension; c++) {
				if (s.masks[mask][r][c]) {
					grid[ir + r][ic + c] = id;
				}
			}
		}
		placed[runeIndex] = true;
	}

	public synchronized void unsetRune(int id, int runeIndex) {
		// clear board of this piece
		for (int r = 0; r < height; r++) {
			for (int c = 0; c < width; c++) {
				if (grid[r][c] == id) grid[r][c] = 0;
			}
		}
		placed[runeIndex] = false;
	}

	public synchronized boolean isBustingPieceCount() {
		int total = 0;
		int freeSpace = 0;
		for (int r = 0; r < height; r++) {
			for (int c = 0; c < width; c++) {
				if (grid[r][c] != 1) freeSpace++;
			}
		}
		for (int i = 0; i < nShape && total <= freeSpace; i++) {
			total += shapes[i].sizeInBlock;
		}
		return total > freeSpace;
	}

	// Icon
	public static int blockSize = 50;

	public synchronized void paint(Graphics g) {
		// center on available space
		int x = (getWidth() - this.getPreferredSize().width) / 2;
		int y = (getHeight() - this.getPreferredSize().height) / 2;
		// grid
		g.setColor(Color.BLACK);
		for (int r = 0; r <= height; r++) {
			g.drawLine(x, y + blockSize * r, x + blockSize * width, y + blockSize * r);
		}
		for (int c = 0; c <= width; c++) {
			g.drawLine(x + blockSize * c, y, x + blockSize * c, y + blockSize * height);
		}

		// cells
		for (int r = 0; r < height; r++) {
			for (int c = 0; c < width; c++) {
				g.setColor(colors[grid[r][c]]);
				g.fillRect(x + blockSize * c + 1, y + blockSize * r + 1, blockSize - 1, blockSize - 1); // -1 & +1 for borders (not sure why it's blockSize-1 instead of -2...
			}
		}
	}

	// MouseListener
	public synchronized void mouseClicked(MouseEvent e) {
		// find out which cell was clicked (ignore click outside, and on borders) (this method is dependant on the paint algorithm)
		int x = e.getX() - ((getWidth() - this.getPreferredSize().width) / 2);
		int y = e.getY() - ((getHeight() - this.getPreferredSize().height) / 2);

		// ignore out of bound
		if (x < 0 || y < 0 || x > blockSize * width || y > blockSize * height) return;

		// ignore border click
		if (x % blockSize == 0 || y % blockSize == 0) return;

		// find cell
		int c = x / blockSize;
		int r = y / blockSize;

		togglePit(r, c);
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
	}

	public void mouseReleased(MouseEvent e) {
	}

}
