/*
 * Written by Rymma from the Phoenix world on April 7th, 2009.
 * Proud Bastoker and member of the BlueIron linkshell.
 * 
 * Shape describes the shape of a rune. 
 * The constructor will compute the rotation masks.
 * 
 * Exemple of shapes:
 * T: xxx   sizeInBlocks = 4, dimension = 3x3, numberOfMasks=4
 *     x
 *     
 * 2x2: xx  sizeInBlocks = 4, dimension = 2x2, numberOfMasks=1
 *      xx
 *      
 * TODO support elemental
 */

package rymma.mmm;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;

import javax.swing.Icon;

public class Shape implements Icon {

	public String name;

	public int dimension;

	public int nMask;

	public boolean masks[][][];

	public int sizeInBlock;

	public Shape(boolean mask[][], int nMask, String name) {
		this.name = name;
		this.dimension = Math.max(mask.length, mask[0].length);
		this.nMask = nMask;
		this.masks = new boolean[nMask][dimension][dimension];
		this.sizeInBlock = 0;
		// copy default mask
		for (int r = 0; r < mask.length; r++) {
			for (int c = 0; c < mask[r].length; c++) {
				this.masks[0][r][c] = mask[r][c];
				// count size in block
				if (mask[r][c]) sizeInBlock++;
			}
		}
		// generate other masks (rotate 90 degree for each mask)
		for (int i = 1; i < nMask; i++) {
			for (int r = 0; r < dimension; r++) {
				for (int c = 0; c < dimension; c++) {
					this.masks[i][r][c] = masks[i - 1][dimension - 1 - c][r];
				}
			}
		}
	}

	public String toString() {
		return name;
	}

	// Icon
	public static int blockSize = 10;

	public static int spacer = 5;

	public int getIconHeight() {
		return (dimension * blockSize) + 2; // +2 padding
	}

	public int getIconWidth() {
		return ((dimension * blockSize) * nMask) + (spacer * (nMask - 1)) + 2; // +2 padding
	}

	public void paintIcon(Component component, Graphics g, int x, int y) {
		// padding
		x++;
		y++;
		// render
		g.setColor(Color.BLACK);
		for (int m = 0; m < nMask; m++) {
			for (int r = 0; r < dimension; r++) {
				for (int c = 0; c < dimension; c++) {
					if (masks[m][r][c]) {
						g.fillRect(x + blockSize * c + (m * (blockSize * dimension + spacer)), y + blockSize * r, blockSize - 1, blockSize - 1);
					}
				}
			}
		}
	}
}
